import hashlib
import os
import tempfile
from contextlib import contextmanager
from pathlib import Path
from zipfile import ZipFile

import git
import pathspec

from efemarai.console import console
from efemarai.definition_checker import DefinitionChecker


class ModelRepository:
    def __init__(self, url=None, branch=None, hash=None, access_token=None):
        self.url = url if url is not None else "."
        self.branch = branch
        self.hash = hash
        self.access_token = access_token
        self.files = self.get_all_files()

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n      url={self.url}"
        res += f"\n      branch={self.branch}"
        res += "\n    )"
        return res

    @property
    def is_remote(self):
        return DefinitionChecker.is_path_remote(self.url)

    def get_all_files(self):
        ignore_files = [
            ignore_file
            for ignore_file in [".gitignore", ".efignore", ".efemaraiignore"]
            if Path.joinpath(Path(self.url), ignore_file).exists()
        ]

        # Do not upload .git folder
        ignore_lines = [".git"]

        # Filter out files based on *ignore specification
        for ignore_file in ignore_files:
            ignore_lines.extend(Path(ignore_file).read_text().splitlines())

        ignore_spec = pathspec.PathSpec.from_lines("gitwildmatch", ignore_lines)

        files = [
            file
            for file in list(Path(self.url).glob("**/*"))
            if not ignore_spec.match_file(str(file))
        ]
        return files

    def calculate_file_hash(self, filepath):
        return ModelHasher._calculate_file_hash(filepath=filepath)

    def get_last_commit(self, access_token):
        remote_heads = git.cmd.Git().ls_remote(
            self.url.replace(
                "//github",
                f"//{access_token}:x-oauth-basic@github",
            ),
            heads=True,
        )
        return remote_heads.split(f"\trefs/heads/{self.branch}")[0]

    @contextmanager
    def archive(self, name):

        with tempfile.TemporaryDirectory() as dirpath:
            zip_name = os.path.join(dirpath, name)

            with ZipFile(zip_name, "w") as zip:
                for file in self.files:
                    zip.write(file)

            filesize_gb = os.stat(zip_name).st_size / (1024**3)
            if filesize_gb > 1:
                console.print(
                    f":face_with_monocle: "
                    f"Model code is suspiciously large ({filesize_gb:.2f} GB) - "
                    f"consider using '.gitignore' or '.efignore' file",
                    style="orange1",
                )

            yield zip_name


class ModelFile:
    def __init__(self, name, url, upload=True, credentials=None, hash_code=None):
        self.name = name
        self.url = url
        self.upload = upload
        self.credentials = credentials
        self.hash_code = hash_code

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n      name={self.name}"
        res += f"\n      url={self.url}"
        res += f"\n      upload={self.upload}"
        res += f"\n      hash_code={self.hash_code}"
        res += f"\n    )"
        return res


class ModelHasher:
    @staticmethod
    def _calculate_file_hash(filepath, hashing_algorithm="sha1"):
        """This function returns the SHA-1 hash
        of the file passed into it."""

        # make a hash object
        if hashing_algorithm == "md5":
            h = hashlib.md5()
        elif hashing_algorithm == "sha1":
            h = hashlib.sha1()

        if str(filepath).startswith("http"):
            return hashlib.sha1(str(filepath).encode("utf-8")).hexdigest()

        # open file for reading in binary mode
        with open(filepath, "rb") as file:
            # loop till the end of the file
            chunk = 0
            while chunk != b"":
                # read only 1024 bytes at a time
                chunk = file.read(1024)
                h.update(chunk)

        # return the hex representation of digest
        return h.hexdigest()


class Model:
    """
    Provides model related functionality.

    It should be created through the :class:`efemarai.project.Project.create_model` method.

    Example:

    .. code-block:: python
        :emphasize-lines: 2

        import efemarai as ef
        ef.Session().project("Name").create_model(...)
    """

    @staticmethod
    def create(project, name, description, repository, files):
        """
        Creates a model.

        You should use :func:`project.create_model` instead.
        """
        if name is None:
            raise ValueError("Missing model name")

        if description is None:
            description = ""

        if repository is None:
            repository = {}

        if files is None:
            files = []

        if not isinstance(repository, ModelRepository):
            repository = ModelRepository(**repository)

        for f in files:
            f["hash_code"] = ModelHasher._calculate_file_hash(filepath=f["url"])

        model_files_hashes = "".join([f["hash_code"] for f in files])

        # Remote repo case
        if repository.is_remote:
            repo_hash = (
                repository.hash
                if repository.hash is not None
                else repository.get_last_commit(repository.access_token)
            )

        # Local repo case
        else:
            repo_hash = "".join(
                [
                    ModelHasher._calculate_file_hash(filepath=i)
                    for i in repository.files
                    if i.is_file()
                ]
            )

        version = hashlib.sha1(
            (repo_hash + model_files_hashes).encode("utf-8")
        ).hexdigest()  # Model hash

        files = [ModelFile(**f) if not isinstance(f, ModelFile) else f for f in files]

        response = project._put(
            f"api/model/undefined/{project.id}",
            json={
                "name": name,
                "description": description,
                "version": version,
                "repository": {
                    "url": repository.url,
                    "branch": repository.branch,
                    "hash": repository.hash,
                    "access_token": repository.access_token,
                },
                "files": [
                    {
                        "name": f.name,
                        "url": f.url,
                        "upload": f.upload,
                        "credentials": f.credentials,
                        "hash": f.hash_code,
                    }
                    for f in files
                ],
            },
        )
        model_id = response["id"]

        if not repository.is_remote:
            base_name = "model_code.zip"
            endpoint = f"api/modelCode/{model_id}/upload"
            with repository.archive(name=base_name) as archive_name:
                project._upload(archive_name, endpoint)

            project._post(endpoint, json={"archive": base_name})

        for f in files:
            if f.upload:
                project._upload(f.url, f"api/model/{model_id}/upload")

        return Model(project, model_id, name, description, version, repository, files)

    @staticmethod
    def push(project, name, description, repository, files, existing_model):
        """
        Pushes a new model version.

        You should use :func:`project.push_model` instead.
        """
        if name is None:
            raise ValueError("Missing model name")

        if description is None:
            description = ""

        if repository is None:
            repository = {}

        if files is None:
            files = []

        if not isinstance(repository, ModelRepository):
            repository = ModelRepository(**repository)

        for f in files:
            f["hash_code"] = ModelHasher._calculate_file_hash(filepath=f["url"])

        model_files_hashes = "".join([f["hash_code"] for f in files])

        if repository.is_remote:
            # Remote repo
            repo_hash = (
                repository.hash
                if repository.hash is not None
                else repository.get_last_commit(repository.access_token)
            )
        else:
            # Local repo
            repo_hash = "".join(
                [
                    ModelHasher._calculate_file_hash(filepath=f)
                    for f in repository.files
                    if f.is_file()
                ]
            )

        version = hashlib.sha1(
            (repo_hash + model_files_hashes).encode("utf-8")
        ).hexdigest()  # Model hash

        # Do not create a new model version if there are no changes
        if version == existing_model.version:
            console.print(
                f":face_with_monocle: "
                f"There are were no changes found in the files. Skipping...",
                style="orange1",
            )
            return existing_model

        files = [ModelFile(**f) if not isinstance(f, ModelFile) else f for f in files]
        files = Model._point_to_existing_model_files(
            new_model_files=files, existing_model=existing_model
        )
        response = project._put(
            f"api/model/undefined/{project.id}",
            json={
                "name": name,
                "description": description,
                "version": version,
                "repository": {
                    "url": repository.url,
                    "branch": repository.branch,
                    "hash": repository.hash,
                    "access_token": repository.access_token,
                },
                "files": [
                    {
                        "name": f.name,
                        "url": f.url,
                        "upload": f.upload,
                        "credentials": f.credentials,
                        "hash": f.hash_code,
                    }
                    for f in files
                ],
            },
        )
        model_id = response["id"]

        if not repository.is_remote:
            base_name = "model_code.zip"
            endpoint = f"api/modelCode/{model_id}/upload"
            with repository.archive(name=base_name) as archive_name:
                project._upload(archive_name, endpoint)

            project._post(endpoint, json={"archive": base_name})

        for f in files:
            if f.upload:
                project._upload(f.url, f"api/model/{model_id}/upload")

        return Model(project, model_id, name, description, version, repository, files)

    def __init__(self, project, id, name, description, version, repository, files):
        self.project = project
        self.id = id
        self.name = name
        self.description = description
        self.version = version
        self.repository = repository
        self.files = files

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  id={self.id}"
        res += f"\n  name={self.name}"
        res += f"\n  description={self.description}"
        res += f"\n  version={self.version}"
        res += f"\n  repository={self.repository}"
        res += f"\n  files={self.files}"
        res += f"\n)"
        return res

    def delete(self, delete_dependants=False):
        """
        Deletes the model.

        You cannot delete an object that is used in a stress test or a baseline
        (delete those first). Deletion cannot be undone.
        """
        self.project._delete(
            f"api/model/{self.id}/{self.project.id}/{delete_dependants}",
        )

    @staticmethod
    def _point_to_existing_model_files(new_model_files, existing_model):
        """
        Checks if the new model files are different from the old ones.
        Points the url to the old ones if they are the same.
        """
        for f in new_model_files:
            if f.hash_code in [f.hash_code for f in existing_model.files]:
                f.url = [
                    file.url
                    for file in existing_model.files
                    if file.hash_code == f.hash_code
                ][0]
                f.upload = False

        return new_model_files

    def versions(self):
        """
        Returns all model versions.
        """
        return self.project._get(f"api/modelVersions/{self.id}/{self.project.id}")



import hashlib
import os
import zipfile
from time import sleep

import boto3
from appdirs import user_data_dir
from botocore.errorfactory import ClientError

from efemarai.console import console
from efemarai.dataset import Dataset
from efemarai.domain import Domain
from efemarai.model import Model, ModelFile, ModelRepository
from efemarai.problem_type import ProblemType
from efemarai.stress_test import StressTest


class Project:
    """
    Provides project related functionality.
    A project is the main building block around which an ML model evaluation is performed.

    A project is an association of a set of ML models, datasets and domains.
    """

    @staticmethod
    def create(
        session, name, description, problem_type, private=False, exists_ok=False
    ):
        """
        Creates a project.

        You should use `ef.Session().create_project(...)` instead.
        """
        if name is None or not ProblemType.has(problem_type):
            raise ValueError(
                f"Project name '{name}' is not specified or problem type '{problem_type}' not recognized."
            )

        existing_project = next((p for p in session.projects if p.name == name), None)
        if existing_project:
            if exists_ok:
                return existing_project
            else:
                raise ValueError(f"Project {name} already exists and exists_ok=False.")

        response = session._put(
            "api/project",
            json={
                "name": name,
                "description": description,
                "problem_type": problem_type,
                "private": private,
            },
        )
        return Project(session, response["id"], name, description, problem_type)

    def __init__(self, session, id, name, description, problem_type):
        self._session = session
        self.id = id
        self.name = name  #: (str) Name of the project.
        self.description = description  #: (str) Description of the project.
        self.problem_type = ProblemType(
            problem_type
        )  #: (:class:`efemarai.problem_type.ProblemType`) Project problem type.

    def __repr__(self):
        res = f"{self.__module__}.{self.__class__.__name__}("
        res += f"\n  id={self.id}"
        res += f"\n  name={self.name}"
        res += f"\n  description={self.description}"
        res += f"\n  problem_type={self.problem_type}"
        res += f"\n)"
        return res

    def _get(self, endpoint, json=None, params=None):
        return self._session._get(endpoint, json, params, self.id)

    def _post(self, endpoint, json=None, params=None):
        return self._session._post(endpoint, json, params, self.id)

    def _put(self, endpoint, json=None, params=None):
        return self._session._put(endpoint, json, params, self.id)

    def _delete(self, endpoint, json=None, params=None):
        return self._session._delete(endpoint, json, params, self.id)

    def _upload(self, from_url, endpoint):
        return self._session._upload(from_url, endpoint, self.id)

    @property
    def models(self):
        """Returns a list of the models associated with the project."""

        return [
            Model(
                project=self,
                id=model["id"],
                name=model["name"],
                description=model.get("description"),
                version=model.get("version"),
                repository=ModelRepository(
                    url=model["repository"]["url"],
                    branch=model["repository"].get("branch"),
                    hash=model["repository"].get("hash"),
                ),
                files=[
                    ModelFile(
                        name=f["name"],
                        url=f["url"],
                        upload=f["upload"],
                        hash_code=f["hash"],
                    )
                    for f in model["files"]
                ],
            )
            for model in self._get(f"api/models/{self.id}")
        ]

    def model(self, model) -> Model:
        """Returns the model specified by the name.

        Args:
            model (str): A model identifier. It can either be the name of the model or its id.

        Returns:
            :class:`efemarai.model.Model`: A model object or `None` if it doesn't exist.
        """

        model = next((m for m in self.models if m.name == model or m.id == model), None)
        return model

    def create_model(
        self,
        name,
        description=None,
        repository=None,
        files=None,
        exists_ok=False,
        **kwargs,
    ) -> Model:
        """Creates a model.

        Args:
            name (str): Name of the model.
            description (str, optional): Free text description.
            repository (dict, optional): Model repository info. See `ef.model.ModelRepository`.
            files (list[dict]): Files needed to load the model e.g. weights or config
                files. See `ef.model.ModelFile`.
            exists_ok (bool, optional): If the current named model exist, should it be
                returned or an exception raised.

        Raises:
            ValueError: If the model exists and `exists_ok` is set to `False`.

        Returns:
            :class:`efemarai.model.Model`: The created model.
        """

        existing_model = next((m for m in self.models if m.name == name), None)
        if existing_model:
            if exists_ok:
                return existing_model
            else:
                raise ValueError(f"Model {name} already exists and exists_ok=False.")

        model = Model.create(self, name, description, repository, files)
        return model

    def push_model(
        self,
        name,
        description=None,
        repository=None,
        files=None,
        **kwargs,
    ) -> Model:
        """Pushes a new model version or creates a model, if there is no such model with this name.

        Args:
            name (str): Name of the model.
            description (str, optional): Free text description.
            repository (dict, optional): Model repository info. See `ef.model.ModelRepository`.
            files (list[dict]): Files needed to load the model e.g. weights or config
                files. See `ef.model.ModelFile`.

        Returns:
            :class:`efemarai.model.Model`: The created model.
        """

        existing_model = next((m for m in self.models if m.name == name), None)
        if existing_model:
            model = Model.push(
                self,
                name,
                description,
                repository,
                files,
                existing_model,
            )
            return model
        else:
            return Model.create(self, name, description, repository, files)

    @property
    def datasets(self):
        """
        Returns a list of the datasets associated with the project.

        Returns:
            list[:class:`efemarai.dataset.Dataset`]: A list of available datasets.
        """

        return [
            Dataset(
                self,
                dataset["id"],
                dataset["name"],
                dataset["format"],
                dataset["stage"],
                dataset["data_url"],
                dataset.get("annotations_url"),
                dataset["states"][-1]["name"],
                Dataset._parse_classes(dataset["classes"]),
            )
            for dataset in self._get(f"api/datasets/{self.id}")
        ]

    def dataset(self, dataset) -> Dataset:
        """
        Returns the dataset specified by the name or id.

        Args:
            dataset (str): A dataset identifier. It can either be the name of the dataset or its id.

        Returns:
            :class:`efemarai.dataset.Dataset`: A dataset object or `None` if it doesn't exist.
        """

        dataset = next(
            (d for d in self.datasets if d.name == dataset or d.id == dataset), None
        )
        return dataset

    def create_dataset(
        self,
        name,
        format,
        stage,
        data_url,
        annotations_url=None,
        credentials=None,
        upload=True,
        num_datapoints=None,
        mask_generation=None,
        exists_ok=False,
        min_asset_size=15,
        **kwargs,
    ) -> Dataset:
        """Creates a dataset.

        Args:
            name (str): A dataset name.
            format (str): A format of the dataset. Possible formats are specified under `ef.dataset.DatasetFormat`.
            stage (str): Dataset stage - one of [`training`, `validation`, `test`].
            data_url (str): A url to fetch data. Can be local or remote.
            annotations_url (str, optional): A url that specifies the annotation file. Can be local or remote.
            credentials (str, optional): If the dataset if remote, Specify the `access token` or `{username}:{password}` to access.
            upload (bool, optional): Should the data be uploaded to Efemarai? This would make evaluation and testing much faster.
            num_datapoints (int, optional): Should a subset of the datapoints be loaded as a dataset? Specify the number of dataponts. Useful for quick testing. Use `None` to load all.
            mask_generation (str): What strategy should be used for mask generation.
            exists_ok (bool, optional): If the current named dataset exist, should it be returned or an exception raised.
            min_asset_size (int, optional): Assets with width and hight smaller than this will be ignored. Default value of 15.

        Raises:
            ValueError: If the dataset exists and `exists_ok` is set to `False`.

        Returns:
            :class:`efemarai.dataset.Dataset`: The created dataset.
        """

        existing_dataset = next((d for d in self.datasets if d.name == name), None)
        if existing_dataset:
            if exists_ok:
                return existing_dataset
            else:
                raise ValueError(f"Dataset {name} already exists and exists_ok=False.")

        dataset = Dataset.create(
            self,
            name,
            format,
            stage,
            data_url,
            annotations_url,
            credentials,
            upload,
            num_datapoints,
            mask_generation,
            min_asset_size,
        )

        return dataset

    @property
    def domains(self):
        """
        Returns a list of the domains associated with the project.

        Returns:
            list [:class:`efemarai.domain.Domain`]: A list of available domains.
        """

        return [
            Domain(
                self,
                domain["id"],
                domain["name"],
                domain["transformations"],
                domain["graph"],
            )
            for domain in self._get(f"api/domains/{self.id}")
        ]

    def domain(self, domain) -> Domain:
        """
        Returns the domain specified by the name or id.

        Args:
            domain (str): A domain identifier. It can either be the name of the domain or its id.

        Returns:
            :class:`efemarai.domain.Domain`: A domain object or `None` if it doesn't exist.
        """

        domain = next(
            (d for d in self.domains if d.name == domain or d.id == domain), None
        )
        return domain

    def create_domain(
        self, name, transformations, graph, exists_ok=False, **kwargs
    ) -> Domain:
        """
        Creates a domain. **The recommended strategy** is to use the UI to build up the domain visually.
        The domain then can be exported to a yaml configuration file for storage and loaded/reused through `ef.Session().load(...)`.

        Args:
            name (str): The name of the domain.
            transformations (dict): A dictionary of the transformations and their params.
            graph (dict): How the transformations are paired.
            exists_ok (bool, optional): If the current named domain exist, should it be returned or an exception raised.

        Raises:
            ValueError: If the domain exists and `exists_ok` is set to `False`.

        Returns:
            :class:`efemarai.domain.Domain`: The created domain.
        """

        existing_domain = next((d for d in self.domains if d.name == name), None)
        if existing_domain:
            if exists_ok:
                return existing_domain
            else:
                raise ValueError(f"Domain {name} already exists and exists_ok=False.")

        domain = Domain.create(self, name, transformations, graph)
        return domain

    @property
    def stress_tests(self):
        """
        Returns a list of the stress tests associated with the project.

        Returns:
            list [:class:`efemarai.stress_test.StressTest`]: A list of available stress tests.
        """

        return [
            StressTest(
                self,
                test["id"],
                test["name"],
                test["model"],
                test["domain"]["id"],
                test["dataset"]["id"],
                test["states"][-1]["name"],
                test["states"][-1].get("message"),
                test["reports"],
            )
            for test in self._get(f"api/getRuns/{self.id}")["objects"]
        ]

    def stress_test(self, test) -> StressTest:
        """
        Returns the stress test specified by the name or id.

        Args:
            test (str): A stress test identifier. It can either be the name of the stress test or its id.

        Returns:
            :class:`efemarai.stress_test.StressTest`: A stress test object or `None` if it doesn't exist.
        """
        test = next(
            (t for t in self.stress_tests if t.name == test or t.id == test), None
        )
        return test

    def create_stress_test(
        self,
        name,
        model,
        domain,
        dataset,
        num_samples=25,
        num_runs=40,
        concurrent_runs=4,
        **kwargs,
    ) -> StressTest:
        """
        Creates a stress test.

        Args:
            model (str, :class:`efemarai.model.Model`): The model to stress test.
            domain (str, :class:`efemarai.domain.Domain`): The domain from which to generate samples.
            dataset (str, :class:`efemarai.dataset.Dataset`): The dataset with which to generate samples.
            num_samples (int, optional): Sample count to generate.
            num_runs (int, optional): Number of searches to run.
            concurrent_runs (int, optional): Number of concurrent searches.

        Returns:
            :class:`efemarai.stress_test.StressTest`: A stress test object.
        """

        test = StressTest.create(
            self,
            name,
            model,
            domain,
            dataset,
            num_samples,
            num_runs,
            concurrent_runs,
        )

        return test

    def delete(self, delete_dependants=False):
        """
        Deletes a project, including the domains, datasets, models, stress tests. Cannot be undone.
        """

        self._delete(f"api/project/{self.id}/{delete_dependants}")

    def export_datasets(
        self,
        test_run_ids: list = None,
        min_score=0.0,
        include_dataset=False,
        path=None,
        unzip=True,
        ignore_cache=False,
        export_format=None,
    ):
        """
        Returns the vulnerabilities dataset associated with the stress tests.

        Args:
            test_run_ids (list[str]): Test runs ids for the datasets.
            min_score (float, optional): Minimum score to select samples.
            include_dataset (bool, optional): If the original dataset used for each of the stress tests should be included.
            path (str, optional): Path to the downloading location.
            unzip (bool, optional): If the zip file should be unzipped.
            ignore_cache (bool, optional): Force regeneration of the dataset by ignoring the cache. May lead to slower subsequent calls.
            export_format (str): The format of the output vulnerabilities dataset.

        Returns:
            str: The filename of the resulting object.
        """

        if export_format is None:
            if self.problem_type == ProblemType.Classification:
                export_format = "imagenet"
            elif self.problem_type == ProblemType.ObjectDetection:
                export_format = "coco"
            elif self.problem_type == ProblemType.InstanceSegmentation:
                export_format = "coco"

        if export_format is None:
            console.print(":poop: Unsupported problem type.", style="red")
            return None

        if not ignore_cache:
            name = "vulnerabilities_dataset"
            name += f"_{export_format}_{include_dataset}_{min_score:.3f}"
            cache_name = os.path.join(path, name)
            if os.path.exists(cache_name) or os.path.exists(cache_name + ".zip"):
                return cache_name

        access = self._post(
            "api/exportDataset",
            json={
                "id": test_run_ids,
                "format": export_format,
                "merge": include_dataset,
                "min_score": min_score,
                "async_download": True,
            },
        )

        s3 = boto3.client(
            "s3",
            aws_access_key_id=access["AccessKeyId"],
            aws_secret_access_key=access["SecretAccessKey"],
            aws_session_token=access["SessionToken"],
            endpoint_url=access["Url"],
        )

        test_run_names = " ,".join(
            [self.stress_test(test_id).name for test_id in test_run_ids]
        )
        with console.status(f"Generating '{test_run_names}' vulnerabilities dataset"):
            while True:
                try:
                    response = s3.head_object(
                        Bucket=access["Bucket"], Key=access["ObjectKey"]
                    )
                    size = response["ContentLength"]
                    break
                except ClientError:
                    sleep(1)

        with self._session._progress_bar() as progress:
            task = progress.add_task("Downloading dataset ", total=float(size))

            def callback(num_bytes):
                return progress.advance(task, num_bytes)

            os.makedirs(path, exist_ok=True)
            filename = os.path.join(path, os.path.basename(access["ObjectKey"]))

            s3.download_file(
                access["Bucket"], access["ObjectKey"], filename, Callback=callback
            )

        if unzip:
            with console.status("Unzipping dataset"):
                dirname = os.path.splitext(filename)[0]
                with zipfile.ZipFile(filename, "r") as f:
                    f.extractall(dirname)

                os.remove(filename)

                filename = dirname

        console.print(
            (
                f":heavy_check_mark: Downloaded '{test_run_names}' "
                f"vulnerabilities dataset to \n  {filename}"
            ),
            style="green",
        )

        return filename

    def download_joint_vulnerabilities_datasets(
        self,
        test_run_ids: list = None,
        min_score=0.0,
        include_dataset=False,
        path=None,
        unzip=True,
        ignore_cache=False,
        export_format=None,
    ):
        """
        Returns a joint dataset of all specified stress tests.

        Args:
            test_run_ids (list[str]): Ids of the test runs to download.
            min_score (float, optional): Minimum score to select samples.
            include_dataset (bool, optional): If the original dataset used for each stress test should be included.
            path (str, optional): Path to the downloading location.
            unzip (bool, optional): If the zip file should be unzipped.
            ignore_cache (bool, optional): Force regeneration of the dataset by ignoring the cache. May lead to slower subsequent calls.
            export_format (str): The format of the output vulnerabilities dataset.

        Returns:
            str: The filename of the resulting object.
        """

        stress_tests = [self.stress_test(test_id) for test_id in test_run_ids]
        for stress_test in stress_tests:
            if not stress_test.finished:
                console.print(
                    (
                        ":warning: Cannot export vulnerabilities "
                        "dataset as stress test is still running"
                    ),
                    style="yellow",
                )
                return None

        if len(set([stress_test.dataset.id for stress_test in stress_tests])) > 1:
            console.print(
                ":poop: Cannot merge different original datasets."
                "Please specify stress tests with the same original dataset.",
                style="red",
            )
            return None

        if path is None:
            path = user_data_dir(appname="efemarai")

        test_run_ids.sort()
        ids_hash = hashlib.sha1("".join(test_run_ids).encode("utf-8")).hexdigest()[:24]
        path = os.path.join(path, ids_hash)

        filename = self.export_datasets(
            test_run_ids=test_run_ids,
            min_score=min_score,
            include_dataset=include_dataset,
            path=path,
            unzip=unzip,
            ignore_cache=ignore_cache,
            export_format=export_format,
        )
        return filename



import os
import re
import tempfile
import urllib
from contextlib import contextmanager

import yaml
from furl import furl

from efemarai.base_checker import BaseChecker
from efemarai.console import console
from efemarai.dataset import DatasetFormat, DatasetStage
from efemarai.problem_type import ProblemType
from efemarai.runtime_checker import RuntimeChecker


class DefinitionChecker(BaseChecker):
    def __init__(self):
        super().__init__()
        self._status = console.status("Checking...")

    @staticmethod
    def is_path_remote(path):
        return urllib.parse.urlparse(path).scheme in ("http", "https")

    def load_definition(self, filename):
        if not os.path.isfile(filename):
            self._error(f"File '{filename}' does not exist")

        with open(filename) as f:
            contents = f.read()
        contents = os.path.expandvars(contents)

        unknown_environment_variables = list(
            re.findall(r"\$\{([a-zA-Z]\w*)\}", contents)
        )
        if unknown_environment_variables:
            for match in unknown_environment_variables:
                self._error(f"Unknown environment variable '{match}' in '{filename}'")

        return yaml.safe_load(contents)

    def check(
        self,
        definition,
        definition_filename=None,
        check_all=False,
        check_project=False,
        check_datasets=False,
        check_models=False,
        check_domains=False,
        check_runtime=False,
    ):
        if definition_filename is None:
            definition_filename = ""

        try:
            self._status.start()
            if "project" in definition or check_project or check_all:
                self.check_project(definition)

            if "datasets" in definition or check_datasets or check_all:
                self.check_datasets(definition)

            if "models" in definition or check_models or check_all:
                # Model runtime information is resolvable only if it is
                # provided in a file called `efemarai.yaml` since this is
                # what the system looks for when loading a model runtime.
                resolvable_runtime = (
                    os.path.basename(definition_filename) == "efemarai.yaml"
                )
                self.check_models(definition, resolvable_runtime, check_runtime)

            if "domains" in definition or check_domains or check_all:
                self.check_domains(definition)
        except AssertionError:
            self._status.stop()
            return False

        self._status.stop()
        return True

    def check_from_file(
        self,
        definition_filename,
        check_all=False,
        check_project=False,
        check_datasets=False,
        check_models=False,
        check_domains=False,
        check_runtime=False,
    ):
        try:
            definition = self.load_definition(definition_filename)
        except AssertionError:
            return False

        return self.check(
            definition,
            check_all=check_all,
            check_project=check_project,
            check_datasets=check_datasets,
            check_models=check_models,
            check_domains=check_domains,
            check_runtime=check_runtime,
            definition_filename=definition_filename,
        )

    def check_project(self, definition):
        self._status.update("Checking project...")

        project = self._get_required_item(definition, "project")

        name = self._get_required_item(project, "name", "project")

        problem_type = self._get_required_item(project, "problem_type", "project")

        if not ProblemType.has(problem_type):
            self._error(f"Unsupported problem type '{problem_type}' (in 'project')")

    def check_datasets(self, definition):
        self._status.update("Checking datasets...")

        datasets = self._get_required_item(definition, "datasets")

        if not isinstance(datasets, list):
            self._error("'datasets' must be an array")

        known_datasets = set()

        for i, dataset in enumerate(datasets):
            parent = f"datasets[{i}]"

            name = self._get_required_item(dataset, "name", parent)

            if name in known_datasets:
                self._error(f"Multiple datasets named '{name}' exist (in 'datasets')")

            known_datasets.add(name)

            format = self._get_required_item(dataset, "format", parent)
            if not DatasetFormat.has(format):
                self._error(f"Unsupported dataset format '{format}' (in '{parent}')")

            stage = self._get_required_item(dataset, "stage", parent)
            if not DatasetStage.has(stage):
                self._error(f"Unsupported dataset stage '{stage}' (in '{parent}')")

            upload = dataset.get("upload", True)

            if upload:
                annotations_url = dataset.get("annotations_url")
                if annotations_url is not None and not os.path.exists(annotations_url):
                    self._error(
                        f"File path '{annotations_url}' does not exist (in '{parent}')"
                    )

                data_url = dataset.get("data_url")
                if data_url is not None and not os.path.exists(data_url):
                    self._error(
                        f"File path '{data_url}' does not exist (in '{parent}')"
                    )

    def check_models(self, definition, resolvable_runtime, check_runtime):
        self._status.update("Checking models...")

        models = self._get_required_item(definition, "models")

        if not isinstance(models, list):
            self._error("'models' must be an array")

        known_models = set()

        for i, model in enumerate(models):
            parent = f"models[{i}]"
            name = self._get_required_item(model, "name", parent)

            if name in known_models:
                self._error(f"Multiple models named '{name}' exist (in 'models')")

            known_models.add(name)

        for i, model in enumerate(models):
            parent = f"models[{i}]"
            self._status.update(f"Checking {parent}...")

            self._check_files(model, parent)

            with self._check_repository(model, parent) as repo_path:
                if model["name"] in {"${model.name}", "$model.name"}:
                    self._warning(f"Skipping default runtime checks (in {parent})")
                    continue

                try:
                    model["runtime"] = self._resolve_runtime(
                        model, repo_path, resolvable_runtime, parent
                    )
                except AssertionError:
                    self._error(f"Unable to find runtime for {parent}")

                if check_runtime:
                    self._status.update(
                        f"Checking {parent}\n  :left_arrow_curving_right: Running model..."
                    )
                    runtime_checker = RuntimeChecker(
                        model,
                        parent,
                        datasets=definition.get("datasets"),
                        project=definition.get("project"),
                        repo_path=repo_path,
                        print_warnings=self._print_warnings,
                    )
                    try:
                        runtime_checker.check()
                    except Exception as e:
                        raise e

    def check_domains(self, definition):
        self._status.update("Checking domains...")

        domains = self._get_required_item(definition, "domains")

        if not isinstance(domains, list):
            self._error("'domains' must be an array")

        known_domains = set()

        for i, domain in enumerate(domains):
            parent = f"domain[{i}]"
            name = self._get_required_item(domain, "name", parent)

            if name in known_domains:
                self._error(f"Multiple models named '{name}' exist (in 'domains')")

            known_domains.add(name)

            _ = self._get_required_item(domain, "transformations", parent)
            _ = self._get_required_item(domain, "graph", parent)

    @contextmanager
    def _check_repository(self, model, parent):
        repository = model.get("repository", {"url": "."})

        repo_parent = parent + ".repository"

        url = self._get_required_item(repository, "url", repo_parent)

        if DefinitionChecker.is_path_remote(url):
            self._status.update(
                f"Checking {parent}\n  :left_arrow_curving_right: Cloning repo..."
            )

            branch = repository.get("branch")
            hash = repository.get("hash")
            access_token = repository.get("access_token")

            if branch is None and hash is None:
                self._error(f"'branch' or 'hash' must be provided (in '{repo_parent}')")

            # Attempt to import GitPython which checks if a 'git' executable
            # is available and raises an ImportError if it isn't.
            try:
                import git
            except ImportError:
                self._warning(
                    "Skipping remote repo check as no 'git' executable is found."
                )
                return

            with tempfile.TemporaryDirectory() as temp_dir:
                try:
                    clone_url = furl(url)
                    clone_url.username = access_token
                    clone_url.password = "x-oauth-basic"

                    repo = git.Repo.clone_from(
                        clone_url.tostr(),
                        temp_dir,
                        branch=branch,
                        depth=1,
                        single_branch=True,
                    )
                except Exception:
                    self._error(
                        f"Unable to clone repository at '{url}' (in '{repo_parent}')"
                    )

                if hash is not None:
                    try:
                        repo.commit(hash)
                    except git.exc.GitError:
                        self._error(
                            f"Commit '{hash}' does not exist (in '{repo_parent}')"
                        )

                yield repo.working_dir
        else:
            if not os.path.isdir(url):
                self._error(
                    f"Repository path '{url}' (in '{repo_parent}') must be a folder."
                )
            yield url

    def _check_files(self, model, parent):
        files = model.get("files", [])
        known_files = set()
        for i, file in enumerate(files):
            file_parent = parent + f".files[{i}]"
            name = self._get_required_item(file, "name", file_parent)

            if name in known_files:
                self._error(f"Multiple files named '{name}' exist (in '{parent}')")

            known_files.add(name)

            url = self._get_required_item(file, "url", parent + f".files[{i}]")

            if file.get("upload", True) and not os.path.exists(url):
                self._error(f"File path '{url}' does not exist (in '{file_parent}')")

    def _resolve_runtime(self, model, repo_path, resolvable_runtime, parent):
        if "runtime" in model:
            if resolvable_runtime:
                return model["runtime"]

            self._warning(
                f"Runtime will not be used - model runtime will be resolved from"
                f" the respective `efemarai.yaml` file (in {parent}.runtime)"
            )

        default_filename = os.path.join(repo_path, "efemarai.yaml")
        if not os.path.exists(default_filename):
            raise AssertionError()

        definition = self.load_definition(default_filename)
        resolved_models = definition.get("models")

        if not resolved_models:
            raise AssertionError()

        for resolved_model in resolved_models:
            if resolved_model["name"] in {
                model["name"],
                "${model.name}",
                "$model.name",
            }:
                runtime = resolved_model.get("runtime")
                if runtime is None:
                    raise AssertionError()
                return runtime

        raise AssertionError()



import glob
import importlib
import os
import re
import signal
import sys
import traceback
from contextlib import contextmanager

import numpy as np
from PIL import Image

from efemarai.base_checker import BaseChecker
from efemarai.console import console


@contextmanager
def time_limit(identifier, seconds=5 * 60):
    # SIGALRM is not implemented in Windows so cannot limit execution time
    if not hasattr(signal, "SIGALRM"):
        yield
        return

    def signal_handler(signum, frame):
        raise TimeoutError(f"{identifier} has timed out (limit {seconds} seconds)")

    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(seconds)

    try:
        yield
    finally:
        signal.alarm(0)


def gpu_available():
    try:

        # Check PyTorch
        import torch

        return torch.cuda.is_available()
    except Exception:
        pass

    try:
        # Check TensorFlow
        import tensorflow as tf

        return tf.test.is_gpu_available()
    except Exception:
        pass

    try:
        # Check Nvidia
        import subprocess

        subprocess.check_output("nvidia-smi")
        return True
    except Exception:
        return False


class RuntimeChecker(BaseChecker):
    def __init__(
        self,
        model,
        parent=None,
        datasets=None,
        project=None,
        repo_path=None,
        print_warnings=True,
    ):
        super().__init__(print_warnings=print_warnings)

        if parent is None:
            parent = ""

        if datasets is None:
            datasets = []

        if repo_path is None:
            repo_path = os.getcwd()

        self._model = model
        self._parent = parent + ".runtime"
        self._datasets = datasets
        self._project = project
        self._repo_path = repo_path

        self._runtime = self._get_required_item(model, "runtime", parent)

        self._variables = self._init_variables()

        self._entrypoints = {
            "load": self._get_entrypoint("load"),
            "predict": self._get_entrypoint("predict"),
        }

        batch = self._runtime.get("batch", {})
        self._batch_size = batch.get("max_size", 1)
        self._image_size = batch.get("image_size")

        if self._image_size is None:
            name = self._runtime["predict"]["entrypoint"]
            self._warning(
                f"Missing field 'image_size' (in '{self._parent}.batch') -"
                f" either provide it or ensure that your '{name}' function works"
                f" with images with different sizes. If all images in your datasets"
                f" have the same size you can ignore this warning."
            )

    def check(self):
        images = self._get_sample_images()

        cwd = os.getcwd()
        os.chdir(self._repo_path)

        try:
            self._load()
            predictions = self._predict(images)

        except TimeoutError as e:
            self._error(f"{e}")
        except Exception:
            console.print(":poop: Runtime Exception", style="red")

            # Drop stack frames of the checker execution
            lines = traceback.format_exc().split("\n")
            message = lines[:1] + lines[7:]

            console.print("\n".join(message)[:-1])
            raise AssertionError()
        finally:
            os.chdir(cwd)

        self._check_predictions(images, predictions)

    def _load(self):
        with time_limit("'load' entrypoint"):
            return self._call("load")

    def _predict(self, images):
        self._variables["datapoint.image"] = images

        with time_limit("'predict' entrypoint"):
            return self._call("predict")

    def _init_variables(self):
        variables = {
            "model.runtime.device": self._get_device(),
        }

        for file in self._model.get("files", []):
            name = f"model.files.{file['name']}.url"
            value = file["url"]
            if file.get("upload", True):
                value = os.path.abspath(value)
            variables[name] = value

        return variables

    def _get_entrypoint(self, name):
        if self._repo_path not in sys.path:
            sys.path.append(self._repo_path)

        definition = self._get_required_item(self._runtime, name, self._parent)
        entrypoint_parent = self._parent + "." + name

        entrypoint = self._get_required_item(
            definition, "entrypoint", entrypoint_parent
        )
        module_path, function_name = entrypoint.split(":")

        if not os.path.exists(
            os.path.join(self._repo_path, f"{module_path.replace('.', '/')}.py")
        ):
            self._error(
                f"File '{module_path}' does not exist (in '{entrypoint_parent}')"
            )

        try:
            module = importlib.import_module(module_path)
        except Exception:
            self._error(
                f"Unable to import '{module_path}' (in '{entrypoint_parent}')",
                print_exception=True,
            )

        try:
            function = getattr(module, function_name)
        except Exception:
            self._error(
                f"Unable to load '{function_name}' (in '{entrypoint_parent}')",
                print_exception=True,
            )

        return function

    def _get_device(self):
        device = self._runtime.get("device", "cpu")
        gpu_requested = device in ["gpu", "cuda"]
        has_gpu = gpu_available()
        if gpu_requested and not has_gpu:
            self._warning(f"GPU not available, checking on CPU (in {self._parent}) ")
        return "cuda" if gpu_requested and has_gpu else "cpu"

    def _get_sample_images(self):
        images = []
        for dataset in self._datasets:
            pattern = os.path.join(dataset["data_url"], "**/*.*")
            for filename in glob.glob(pattern):
                try:
                    image = Image.open(filename)
                    if self._image_size is not None:
                        image.resize(self._image_size)
                    images.append(image)

                    if len(images) == self._batch_size:
                        return images
                except IOError:
                    continue

        if images:
            return images

        for _ in range(self._batch_size):
            data = np.random.rand(400, 400, 3) * 255
            images.append(Image.fromarray(data.astype(np.uint8)))
        return images

    def _call(self, function_name):
        function = self._entrypoints[function_name]

        args = {
            arg["name"]: self._resolve_variable(arg["value"])
            for arg in self._runtime[function_name].get("inputs", [])
        }

        self._save_inputs(function_name, args)

        result = function(**args)
        self._save_output(function_name, result)

        return result

    def _save_inputs(self, function_name, args):
        for name, value in args.items():
            self._variables[f"model.runtime.{function_name}.inputs.{name}"] = value

    def _save_output(self, function_name, result):
        output = self._runtime[function_name]["output"]
        name = output["name"]
        self._variables[f"model.runtime.{function_name}.output.{name}"] = result

    def _resolve_variable(self, value):
        if not (isinstance(value, str) and value.startswith("$")):
            return value

        match = re.match(r"\$\{?(?P<varname>[.\w]+)\}?", value)
        varname = match.group("varname")

        if varname not in self._variables:
            self._error(f"Unknown runtime variable '{varname}' (in '{self._parent}')")

        return self._variables[varname]

    def _check_predictions(self, images, predictions):
        if not isinstance(predictions, list):
            self._error(
                f"Predict function must return a {type([])},"
                f" but got {type(predictions)} (in '{self._parent}')"
            )

        if len(images) != len(predictions):
            self._error(
                f"Invalid number of predictions: expected {len(images)},"
                f" but got {len(predictions)} (in '{self._parent}')"
            )

        for image, prediction in zip(images, predictions):
            self._check_prediction(image, prediction)

    def _check_prediction(self, image, prediction):
        if not isinstance(prediction, dict):
            self._error(
                f"Each prediction must be a dict,"
                f" not {type(prediction)} (in '{self._parent})"
            )
        supported_keys = {
            "classes",
            "scores",
            "boxes",
            "masks",
            "segmentation",
            "regressed_values",
            "image_output",
            "keypoints",
        }
        actual_keys = set(prediction.keys())

        unsupported_keys = actual_keys - supported_keys
        if unsupported_keys:
            self._error(
                f"Unsupported prediction keys: {unsupported_keys}"
                f" (in '{self._parent}.predict')",
            )

        if self._project is not None:
            expected_keys = {
                "Classification": {"classes", "scores"},
                "ObjectDetection": {"boxes", "classes", "scores"},
                "InstanceSegmentation": {"masks", "boxes", "classes", "scores"},
                "SemanticSegmentation": {"segmentation"},
                "ImageRegression": {"regressed_values"},
                "Keypoints": {"keypoints"},
            }[self._project["problem_type"]]

            missing_keys = expected_keys - actual_keys
            if missing_keys:
                self._error(
                    f"Missing prediction keys: {missing_keys}"
                    f" (in '{self._parent}.predict')"
                )

        self._check_classes(prediction)
        self._check_masks(prediction)
        self._check_boxes(prediction)
        self._check_segmentation(prediction)

    def _check_classes(self, prediction):
        classes = prediction.get("classes")

        if classes is None:
            return

        if not all(isinstance(cls, int) for cls in classes):
            self._error(
                f"Predicted 'classes' must contain class indices and"
                rf" so be of type 'List\[int]' (in '{self._parent}.predict')"
            )

    def _check_masks(self, prediction):
        masks = prediction.get("masks")

        if masks is None:
            return

        if not isinstance(masks, list):
            self._error(
                f"Predicted masks per image must be a '{type([])}',"
                f" not '{type(masks)}' (in '{self._parent}.predict')"
            )

        if "classes" not in prediction:
            self._error("Missing 'classes' for predicted 'masks'")

        classes = prediction["classes"]

        if not isinstance(classes, list):
            self._error(
                f"Predicted masks classes per image must be a '{type([])}',"
                f" not '{type(classes)}' (in '{self._parent}.predict')"
            )

        if len(classes) != len(masks):
            self._error(
                f"Invalid prediction: got {len(classes)} class labels,"
                f"  but {len(masks)} masks (in '{self._parent}.predict')"
            )

        for mask in masks:
            mask_shape = np.asarray(mask).shape
            if len(mask_shape) != 2:
                self._error(
                    f"Invalid mask shape: {mask_shape}, "
                    f" expected mask shape is '[width, height]'"
                    f" (in '{self._parent}.predict')"
                )

            if np.min(mask) < 0 or np.max(mask) > 1:
                self._error(
                    f"Expected mask range is [0; 1], not"
                    f" [{np.min(mask)}; {np.max(mask)}] "
                    f" (in '{self._parent}.predict')"
                )

    def _check_boxes(self, prediction):
        boxes = prediction.get("boxes")

        if boxes is None:
            return

        if not isinstance(boxes, list):
            self._error(
                f"Predicted boxes per image must be a '{type([])}',"
                f" not '{type(boxes)}' (in '{self._parent}.predict')"
            )

        if "classes" not in prediction:
            self._error("Missing 'classes' for predicted 'boxes'")

        classes = prediction["classes"]

        if not isinstance(classes, list):
            self._error(
                f"Predicted boxes classes per image must be a '{type([])}',"
                f" not '{type(classes)}' (in '{self._parent}.predict')"
            )

        if len(classes) != len(boxes):
            self._error(
                f"Invalid prediction: got {len(classes)} class labels,"
                f" but {len(boxes)} boxes (in '{self._parent}.predict')"
            )

        for box in boxes:
            x1, y1, x2, y2 = box
            if x1 >= x2 or y1 >= y2:
                self._error(
                    f"Invalid box coordinates: {box}, expected box format is"
                    f" [x1, y1, x2, y2] where (x1; y1) is the top-left corner"
                    f" and (x2; y2) is the bottom-right corner"
                    f" (in '{self._parent}.predict')"
                )

    def _check_segmentation(self, prediction):
        segmentation = prediction.get("segmentation")
        if segmentation is None:
            return

        segmentation_shape = np.asarray(segmentation).shape

        if len(segmentation_shape) != 3 or segmentation_shape[0] != 1:
            self._error(
                f"Invalid segmentation shape: {segmentation_shape}, "
                f" expected segmentation shape is '[1, width, height]'"
                f" (in '{self._parent}.predict')"
            )



from enum import Enum


class JobState(Enum):
    """
    Possible Test Run, Baseline and Dataset States.
    """

    NotStarted = "NotStarted"
    Starting = "Starting"
    Evaluating = "Evaluating"
    Testing = "Testing"
    Loading = "Loading"
    DataLoaded = "Data Loaded"
    CalculatingMetadata = "Calculating Metadata"
    GeneratingReport = "Generating Report"
    Finished = "Finished"
    Failed = "Failed"
    Stopped = "Stopped"
